------------------------------------------------------------------------------
-- Simulator for Elliott 900 series 18 bit Machines. -------------------------
------------------------------------------------------------------------------
-- Terry Froggatt, 15 December 1992, Meridian Ada.   --
-- Don Hunter reduced the store size to 16384 (MEMORY_INDEX)
-- on 30-DEC-1994.
-- Don Hunter made the input come from EIN.DAT in local
-- directory on 25-MAR-1996.
-- And filled the store initially from M0.DAT and M1.DAT
-- on 10-APR-1996.
-- Also made OPERAND_ADDRESS mod 65536.
-- 18-APR-1996 Made it accept RLB stuff.
-- 21-APR-1996 Cleared Q register on a "/" instruction.
-- 02-MAY-1996 Check for excessive execution of 15 7168 on a faulty program.
-- 18-MAY-1996 Renamed to SIM900AL.
-- 27-SEP-1997 Amended to do library operations.
-- 05-APR-1998 Amended to compile after library installation
--             and correct "12" instruction multiplying by
--             "-1.0".
-- 29-AUG-1998 Change constant 242144 to 262144 and amend code too !
-- 10-SEP-2000 Made it recognise 15 4864 plotter instruction
--  3-MAR-2008 Erik Baigar, Transition to GTKada-1.12 and gnat-3.15 started:
--              5-MAR: Graphics stuff disabled and related packages
--                     eliminated, ToDo-Stuff marked by EBRM-0305
--             10-MAR: Unification of all integers types -> integer
--             11-MAR: Elimination of the ARRAY_OBJECT package
--             11-MAR: BIT_OPS from gant's system used instead of the
--                     Meridian Bit_Ops.
--             11-MAR: Integer-IO adapted to the needs of gnat.
--             13-MAR: CR and LF characters for file File read implemented,
--                     uups, only works with the original setting...
--             14-MAR: Implemented basic gtk-binding based on the
--                     scribbling example supplied with GtkAda-1.2.
--             15-MAR: Added instruction counter and time elapsed message.
-- 15-MAR-2008 Erik Baigar: Optimization of GTK-Version: Event-Handler,
-- V1.0        Draw-Sceduler, kill Button - 1st try on Solaris/SPARC,
--             encountered problems with CRLF-Translation...
-- 01-MAY-2010 Added Raw-Patch from Terry Froggatt - Suggested Comment from TJF
-- V1.1        -------------------------------------------------------
--             -- Terry Froggatt, 15 December 1992, Meridian Ada,   --
--             -- with raw binary input 29 April 2010.              --
--             -------------------------------------------------------
--             * New command IPRAW to simulate reading raw tapes via files.
--             * Extended culti script for automatic generation of binary distri-
--               butions for the platform executed on [in /tmp for irix, solaris
--               and in C:\temp on Windows].
--             * Adapted patch for different gnat byte-io via definition
--             * Slightly improved Messages, output of runtime fixed
-- 04-MAY-2010 EB: Added trap for EHB writing to paper tape: Compiled tapes
--             are sent with operand 6144 whereas output goes to 4168. Added
--             new command CTAPE "filename" in analogy to Terry's patch
--             above using the SEQUENTIAL_IO -> Works very nice and successfully
--             compiled all versions of SAP for 102 using EHB. Automatic
--             handling of the current version implemented in make scripts and
--             Ada source code. Edit Version numer ONLY later in this file's
--             definition of VERSION and DATE as constant string. 5-MAY-2010.
-- 02-SEP-2010 EB: Slightly modified and less diverted output. V1.1a
-- 08-OCT-2010 EB: Added readme files, readme.1st, retrofitted Don Hunters
--             daytime patch from 09-JAN-2003 'Made it output date and time
--             for a "15 4867" instruction', switched to new Algol libs
--             and added more examples. Asked Don for permission to do so,
--             no answer until 1/2/2011.  :-(
--  1-Feb-2011 EB: Updated multiplication instruction (function 12) according
--             to recommendation Terry Froggatt has given in his paper on the
--             multiply bug on 15th of November 2010.
-- V1.2        -------------------------------------------------------
-- 26-APR-2011 EB: Added Messages for TROUBLE-Exits -> Just to be polite
--  5-MAY-2011 EB: Fixed problem preventing SIM (simulator for 12 bit
--             Elliott 102/905 machines) from working within the sim900:
--             Handling of 15 7168 was not correctly implemented - there
--             exists a shadow register in the real machine which changes
--             behaviour obeying this instruction in base level (4)!
--  6-MAY-2011 EB: Implemented directive DUMP102 for dumping the virtual
--             registers of the machine simulated by SIM. This gives only
--             correct output if SIM is loaded and put into reset state.
-- V1.3        -------------------------------------------------------
------------------------------------------------------------------------------

with TEXT_IO;
use  TEXT_IO;

with system.bit_ops;      -- Bit operations, gnat specific, EB, 3-MAR-2008
use system.bit_ops;       -- Unfortunately they work different than Meridian's

with Ada.Integer_Text_IO; -- New Integer-IO-routines required for reading
                          -- and writing numbers from files ot the
                          -- console, EB, 11-MAR-2008

with Ada.Long_Integer_Text_IO; -- New Integer-IO-routines for printing
                               -- Instruction Counter, 14-MAR-2008, EB

with Ada.Calendar;        -- Make clock avaliable for a time elapsed display,
use Ada.Calendar;         -- very useful for GUI optimization, EB, 14-MAR-2008

with Ada.Strings.Fixed;   -- Functions used for triming strings, e.g. to avoid
use Ada.Strings.Fixed;    -- leading spaces, EB, 6-MAY-2011

with SEQUENTIAL_IO;       -- Needed for Terry's Raw-Input-Patch, 1-MAY-2010
                          -- and also for my patch for tape output CTAPE, 4-MAY-2010
------------------------------------------------------------------------------
-- Use Gtk-Ada 1.2 to do the graphics part of the Plotter-Emulation
with Glib;             use Glib;
-- Gdk
with Gdk;
with Gdk.Drawable;     use Gdk.Drawable;
with Gdk.Event;        use Gdk.Event;
with Gdk.Pixmap;       use Gdk.Pixmap;
with Gdk.Rectangle;    use Gdk.Rectangle;
with Gdk.Types;        use Gdk.Types;
with Gdk.Window;       use Gdk.Window;
-- Gtk
with Gtk;              use Gtk;
with Gtk.Box;          use Gtk.Box;
with Gtk.Button;       use Gtk.Button;
with Gtk.Drawing_Area; use Gtk.Drawing_Area;
with Gtk.Enums;        use Gtk.Enums;
with Gtk.Main;         use Gtk.Main;
with Gtk.Handlers;     use Gtk.Handlers;
with Gtk.Style;        use Gtk.Style;
with Gtk.Window;       use Gtk.Window;
-------------------------------------------------------------------------------

procedure Sim900al_nt_r is

   ----------------------------------------------------------------------------
   -- Edit Version here, but strictly keep line format as it is processed by
   -- the automatic make scripts for all platforms, 5-MAY-2010, EB
   ----------------------------------------------------------------------------
   VERSION           : constant string := "1.3";
   DATE              : constant String := "29-MAY-2011";
   ----------------------------------------------------------------------------
   Pixmap                     : Gdk_Pixmap;
   Window                     : Gtk_Window;
   GUIquit,WAITquit           : Boolean := False;
   DummyMain                  : Boolean;
   PicChanged                 : Boolean;
   InstCount                  : Long_Integer := 0 ;
   SimStart, SimStop          : TIME;
   Month,Day,Year             : Integer;
   Hour,Minute,Second         : Integer;
   Start,Stop                 : Day_Duration;
   -- Added for the DUMP102 facility, 5-MAY-2011, EB
   s102,a102                  : Integer;
   i102,i1021,i1022,i1023     : Integer;
   B102,D102                  : Integer;

   -- Adaption of Terry's raw-read-patch to gnat-Ada
   type BYTE_INTEGER  is mod 256;

   -- Part of Terry's raw input patch, 1-MAY-2010:
   package BYTE_IO is new SEQUENTIAL_IO (BYTE_INTEGER);
   FILE: BYTE_IO.FILE_TYPE;
   BYTE: BYTE_INTEGER;

   -- Patch to trap output of EHB (device 6144) to tape-file, 5-MAY-2010, EB
   TAPE: BYTE_IO.FILE_TYPE;

   FATAL_EXCEPTION, TROUBLE: exception;
   BATCH_MODE_RATHER_THAN_INTERACTIVE: BOOLEAN := TRUE;

   -- The CRLF characters have to be different for Windows/UNIX, 13-MAR-2008, EB
   -- but for the Elliott-use we always use "LF only", it's OK.
    CRLF: constant CHARACTER := CHARACTER'VAL(10); -- ASCII.LF, Windows
    HALT: constant CHARACTER := CHARACTER'VAL(20); -- ASCII.DC4.

   ---------------------- lot of GTK-Stuff here, 14-MAR-2008, EB ----------------

   package Configured is new Gtk.Handlers.Callback
     (Widget_Type => Gtk_Drawing_Area_Record);
   package Destroyed is new Gtk.Handlers.Callback
     (Widget_Type => Gtk_Window_Record);

   ---------------------
   -- Configure_Event --
   ---------------------

   procedure Configure_Event (Drawing_Area : access Gtk_Drawing_Area_Record'Class) is
      Win: GDK_Window;
      Width: Gint;
      Height: Gint;

   begin
        Win:=Get_Window (Drawing_Area);
        Get_Size (Win, Width, Height);
        GDK.Pixmap.Gdk_New (Pixmap, Win, Width, Height, -1);
        Draw_Rectangle (Pixmap, Get_White(Get_Style(Drawing_Area)),
                        True, 0, 0, Width, Height);
   end Configure_Event;

   ------------------
   -- Expose_Event --
   ------------------

   procedure Expose_Event (Drawing_Area : access Gtk_Drawing_Area_Record'Class;
                           Event: in Gdk.Event.Gdk_Event)
   is
      Area : Gdk_Rectangle := Get_Area (Event);
   begin
      Draw_Pixmap (Get_Window(Drawing_Area),
                   Get_FG_GC(Get_style(Drawing_Area),State_Normal),
                   Pixmap, Area.X, Area.Y, Area.X, Area.Y,
                   Gint(Area.Width), Gint(Area.Height));
   end Expose_Event;

   ----------------
   -- Draw_Brush --
   ----------------

   procedure Draw_Brush (Widget : access Gtk_Drawing_Area_Record'Class;
                         X,Y: Gint) is
   begin
      Draw_Arc (Pixmap, Get_Black(Get_Style(Widget)), False,
                x-15, y-15, 2*15, 2*15, 0, 360*64);
      Draw (Widget);
   end Draw_Brush;

   ------------------------
   -- Button_Press_Event --
   ------------------------

   procedure Button_Press_Event
     (Widget: access Gtk_Drawing_Area_Record'Class;
      Event: in  Gdk_Event)
   is
      use type Gdk.Pixmap.Gdk_Pixmap;
   begin
      if Get_Button(Event) = 1 and pixmap /= Null_Pixmap then
         Draw_Brush (Widget, Gint(Get_X(Event)), Gint(Get_Y(Event)));
         PicChanged:= True; -- Request Redraw
      end if;
   end Button_Press_Event;

   -------------------------
   -- Motion_Notify_Event --
   -------------------------

   procedure Motion_Notify_Event
     (Widget: access Gtk_Drawing_Area_Record'Class; Event: in Gdk_Event)
   is
      X,Y: Gint;
      State: Gdk_Modifier_Type;
      Result: Gdk_Window;

      use type Gdk.Pixmap.Gdk_Pixmap;
   begin
      if Get_Is_Hint (Event) then
         Get_Pointer (Get_Window(Widget),X,Y,State,Result);
      else
         x := Gint(Get_X(Event));
         y := Gint(Get_Y(Event));
         State := Get_State (Event);
      end if;

      if (State and BUTTON1_MASK) /= 0 and Pixmap /= Null_Pixmap then
        Draw_Brush (Widget, X, Y);
      end if;
   end Motion_Notify_Event;

   ---------
   -- Bye --
   ---------

   procedure Bye (Window : access Gtk.Window.Gtk_Window_Record'Class) is
      pragma Warnings (Off, Window);
   begin
      -- Used to quit and close window after simulation completed
      if WAITquit then Gtk.Main.Main_Quit; end if;
      -- Used to raise fatal exception during simulation running
      GUIquit:= True;
   end Bye;

   -- Used GUI-Graphics elements:

   Button : Gtk_Button;
   Vbox : Gtk_Box;
   Drawing_Area: Gtk_Drawing_Area;

---------------------------End of GTK-Stuff ----------------------------------





--------------------------- Give nice date output  ---------------------------
function Month_Name(Month : Month_Number) return String is
   type Months is (JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT,
                        NOV, DEC);
begin
   -- convert month number to enumeration value then convert enumeration
   -- value to a string
   return Months'Image(Months'Val(Month - 1));
end Month_Name;
---------------------------End of month conversion ---------------------------



--
-- New declarations on 25-SEP-1997
--
  lib: file_type;
  library: constant string := "Sim900al-NT.LIBRARY.dat";
  lib_flag1: boolean := FALSE;
  lib_flag2: boolean := FALSE;
  has_run:   boolean := FALSE;
--

  f: file_type;
  ein: constant string := "ElliottIn.dat";
  MODULE: file_type;
  m0: constant string := "Sim900al-NT.M0.dat";
  m1: constant string := "Sim900al-NT.M1.dat";
  top, bot, int_level: integer;
  run_after_lib: integer;

-- New declarations on 10-SEP-2000, EB
  down : boolean := FALSE;
  plot : boolean := FALSE;
  x, x1, x2 : integer := 320;
  y, y1, y2 : integer := 160;

  INPUT_BUFFER: array (0..255) of CHARACTER;
  INPUT_COLUMN: INTEGER := -1;
  OUTPUT_COLUMN: INTEGER := 0;

  type INPUT_OUTPUT_MODES is
    (ASCII_MODE, BINARY_MODE, LEGIBLE_MODE, SINK_MODE);
  INPUT_MODE, OUTPUT_MODE: INPUT_OUTPUT_MODES := BINARY_MODE;

  type OP_KIND is
    (TRANSLATE, RUN, ADD_RLB);
  OP: OP_KIND := TRANSLATE;

  TRACING_ON, AMENDING_ON: BOOLEAN := FALSE;
  AMEND_ADDRESS: INTEGER := 8;
  AMEND_WORD, AMEND_FIELD: INTEGER;
  DISPLAY_FROM, DISPLAY_TO: INTEGER;
  BREAKPOINT_ADDRESS: INTEGER := -1;

  type MACHINE_STATES is
    (UNDEFINED_STATE, RUNNABLE_STATE, INPUT_HUNG_STATE);
  MACHINE_STATE: MACHINE_STATES := UNDEFINED_STATE;

  ACCUMULATOR:                   INTEGER := 0;
  Q_REGISTER:                    INTEGER := 0;
  B_REGISTER_ADDRESS:            INTEGER := 1;
  SEQUENCE_CONTROL_REGISTER:     INTEGER := 0;
  OLD_SEQUENCE_CONTROL_REGISTER: INTEGER := 0;
  INSTRUCTION:                   INTEGER := 0;
  OPERAND_ADDRESS:               INTEGER := 0;
  MEMORY_BUFFER:                 INTEGER := 0;
  X_BITS:                        INTEGER := 0;
  PERIPHERAL_DATA:               INTEGER := 0;

  -- The next line had 0..65535 in it before 30-DEC-1994
  -- And it has again on 11-MAY-1996
  -- Replaced array package by direct array, EB, 12-MAR-2008
  subtype MEMORY_INDEX is INTEGER range 0..65535;
  MEMORY: array (MEMORY_INDEX) of INTEGER;


  -- Integer increment is defined already -> disabled, EB, 12-MAR-2008
  -- procedure INCREMENT ( I: in out INTEGER ) is
  -- begin
  --   I := INTEGER'SUCC (I);
  -- end INCREMENT;

  procedure INCREMENT ( I: in out INTEGER ) is
  begin
    I := INTEGER'SUCC (I);
  end INCREMENT;

  procedure CHARACTER_PUT ( ITEM: CHARACTER ) is
  begin
    if ITEM = CRLF
    then NEW_LINE; OUTPUT_COLUMN := 0;
    else PUT (ITEM=>ITEM); INCREMENT (OUTPUT_COLUMN);
    end if;
  end CHARACTER_PUT;

  procedure ENSURE_NEW_LINE is
  begin
    if OUTPUT_COLUMN > 0 then CHARACTER_PUT (CRLF); end if;
  end ENSURE_NEW_LINE;

  procedure ENSURE_NEW_FIELD is
  begin
    loop
      CHARACTER_PUT (' '); -- Always at least one.
      exit when OUTPUT_COLUMN mod 15 = 0;
    end loop;
  end ENSURE_NEW_FIELD;

  procedure STRING_PUT ( ITEM: STRING ) is
  begin
    for I in ITEM'RANGE loop CHARACTER_PUT ( ITEM(I) ); end loop;
  end STRING_PUT;

  procedure MESSAGE_PUT ( ITEM: STRING ) is
  begin
    ENSURE_NEW_LINE;
    STRING_PUT ("Sim900al-NT-R: ");
    STRING_PUT (ITEM);
    ENSURE_NEW_LINE;
  end MESSAGE_PUT;

  procedure ERROR_PUT ( ITEM: STRING ) is
  begin
    MESSAGE_PUT (ITEM);
    if BATCH_MODE_RATHER_THAN_INTERACTIVE then raise FATAL_EXCEPTION; end if;
  end ERROR_PUT;

  procedure NATURAL_PUT ( ITEM: INTEGER ) is
  begin
    if ITEM >= 10 then NATURAL_PUT (ITEM / 10); end if;
    CHARACTER_PUT ( CHARACTER'VAL ( INTEGER(ITEM rem 10) + CHARACTER'POS('0') ) );
  end NATURAL_PUT;

  procedure WORD_PUT ( NAME: CHARACTER; ITEM: INTEGER ) is
  begin
    ENSURE_NEW_FIELD;
    if NAME /= ' ' then
      CHARACTER_PUT (NAME); CHARACTER_PUT (' ');
      CHARACTER_PUT ('='); CHARACTER_PUT (' ');
    end if;
    if ITEM >= 131072
    then CHARACTER_PUT ('-'); NATURAL_PUT (262144 - ITEM);
    else CHARACTER_PUT ('+'); NATURAL_PUT (ITEM);
    end if;
  end WORD_PUT;

  procedure INFORMATION_PUT (
    ADDRESS: INTEGER;
    DISPLAY_REGISTERS_RATHER_THAN_FORMATS: BOOLEAN )
  is
    WORD: INTEGER; -- Not constant.
  begin
    if ADDRESS > 16383 then
                            MESSAGE_PUT ( "Adress not fitting 14 bits encountered! Dumping:" );
                            raise TROUBLE;
    end if;
    WORD := MEMORY(ADDRESS);
    ENSURE_NEW_LINE;
    CHARACTER_PUT ('('); NATURAL_PUT (ADDRESS mod 8192);
    if ADDRESS >= 8192 then
      CHARACTER_PUT ('^'); NATURAL_PUT (ADDRESS / 8192);
    end if;
    CHARACTER_PUT (')');
    ENSURE_NEW_FIELD;
    if WORD >= 131072 then CHARACTER_PUT ('/'); end if;
    NATURAL_PUT ( (WORD / 8192) mod 16 ); CHARACTER_PUT (' ');
    NATURAL_PUT (WORD mod 8192); CHARACTER_PUT (';');
    if DISPLAY_REGISTERS_RATHER_THAN_FORMATS then
      WORD_PUT ( 'A', ACCUMULATOR );
      WORD_PUT ( 'Q', Q_REGISTER );
      WORD_PUT ( 'B', MEMORY (B_REGISTER_ADDRESS) );
    else
      WORD_PUT ( ' ', WORD );
      ENSURE_NEW_FIELD;
      CHARACTER_PUT ('&');
      for I in 1..6 loop -- OCTAL_PUT unsigned with leading zeros.
        CHARACTER_PUT (
          CHARACTER'VAL( (WORD / 32768) + CHARACTER'POS('0') ) );
        WORD := (WORD * 8) mod 262144;
      end loop;
      WORD := MEMORY (ADDRESS);
      ENSURE_NEW_FIELD;
      CHARACTER_PUT ('#');
      while WORD /= 0 loop -- ALPHANUMERIC_PUT.
        CHARACTER_PUT (
          CHARACTER'VAL( (WORD / 4096) + CHARACTER'POS(' ') ) );
        WORD := (WORD * 64) mod 262144;
      end loop;
    end if;
    ENSURE_NEW_LINE;
  end INFORMATION_PUT;

  function CHARACTER_LOOK return CHARACTER is
  begin
    if INPUT_COLUMN = -1 then
      if BATCH_MODE_RATHER_THAN_INTERACTIVE
-- 02-MAY-1996 was ENSURE_NEW_LINE instead of null;
      then null; else MESSAGE_PUT ( "What Next?" ); end if;
      loop
        INCREMENT (INPUT_COLUMN);
        if END_OF_LINE then
           begin SKIP_LINE;
            -- Put("!EOL!");
            INPUT_BUFFER (INPUT_COLUMN) := CRLF;
          exception when END_ERROR =>
            INPUT_BUFFER (INPUT_COLUMN) := HALT;
          end; exit;
        end if;
        GET (ITEM => INPUT_BUFFER (INPUT_COLUMN) );
      end loop;
      INPUT_COLUMN := 0;
    end if;
    return INPUT_BUFFER (INPUT_COLUMN);
  end CHARACTER_LOOK;

  procedure CHARACTER_GET is
  begin
    if CHARACTER_LOOK = HALT then
      BATCH_MODE_RATHER_THAN_INTERACTIVE := TRUE;
      ERROR_PUT ( "Read Beyond File" );
    elsif CHARACTER_LOOK = CRLF then
      INPUT_COLUMN := -1; -- But don't read the line yet,
      -- as this might give trouble in interactive mode.
    else
      INCREMENT (INPUT_COLUMN);
    end if;
  end CHARACTER_GET;

  function CHARACTER_MATCH ( ITEM: CHARACTER ) return BOOLEAN is
  begin
    if CHARACTER_LOOK = ITEM
    then CHARACTER_GET; return TRUE;
    else return FALSE;
    end if;
  end CHARACTER_MATCH;

  function STRING_MATCH ( ITEM: STRING ) return BOOLEAN is
    -- Where ITEM must be in upper-case.
    -- Could produce a version (for Perspective)
    -- where ITEM is mainly in lower-case but
    -- with the abbreviation letters in upper-case.
    BACKTRACK_COLUMN: INTEGER := INPUT_COLUMN;
    DIFFERENCE_IN_POS: INTEGER;
  begin
    for I in ITEM'RANGE loop
      DIFFERENCE_IN_POS :=
        CHARACTER'POS(CHARACTER_LOOK) - CHARACTER'POS(ITEM(I));
      if DIFFERENCE_IN_POS /= 0 and then DIFFERENCE_IN_POS /= 32
      then
        if BACKTRACK_COLUMN = -1 then INPUT_COLUMN := 0;
        else INPUT_COLUMN := BACKTRACK_COLUMN; end if;
        return FALSE;
      end if;
      CHARACTER_GET;
    end loop;
    return TRUE;
  end STRING_MATCH;

  function NATURAL_GET ( BASE: INTEGER := 10 ) return INTEGER is
    ITEM: INTEGER := 0;
    DIGIT: INTEGER;
    ERROR: BOOLEAN := TRUE;
  begin
    loop
      DIGIT := INTEGER ( CHARACTER'POS(CHARACTER_LOOK) - CHARACTER'POS('0') );
      exit when DIGIT not in 0..BASE-1;
      ITEM := ITEM * BASE + DIGIT; CHARACTER_GET; ERROR := FALSE;
    end loop;
    if ERROR then ERROR_PUT ( "Bad Parameter" ); end if;
    return ITEM;
  end NATURAL_GET;

  function PARAMETER_GET return INTEGER is
  begin
    while CHARACTER_LOOK <= ' ' loop CHARACTER_GET; end loop;
    return NATURAL_GET;
  end PARAMETER_GET;

  procedure IGNORE_COMMENTS is
  begin
    loop
      if CHARACTER_MATCH ('(') then IGNORE_COMMENTS;
      elsif CHARACTER_MATCH (')') then exit;
      else CHARACTER_GET; end if;
    end loop;
  end IGNORE_COMMENTS;

  procedure CLEAR_STORE ( LAST: INTEGER := MEMORY_INDEX'LAST ) is
  begin
    for MEMORY_ADDRESS in MEMORY_INDEX'FIRST..LAST loop
      MEMORY(MEMORY_ADDRESS) := 0;
    end loop;
  end CLEAR_STORE;

  procedure ESTABLISH_INITIAL_INSTRUCTIONS is
  begin
    MEMORY(8180):= 262144 - 3;
    MEMORY(8181):= 8192 *  0 + 8180;
    MEMORY(8182):= 8192 *  4 + 8189;
    MEMORY(8183):= 8192 * 15 + 2048;
    MEMORY(8184):= 8192 *  9 + 8186;
    MEMORY(8185):= 8192 *  8 + 8183;
    MEMORY(8186):= 8192 * 15 + 2048;
    MEMORY(8187):= 8192 *  5 + 8180 + 131072;
    MEMORY(8188):= 8192 * 10 +    1;
    MEMORY(8189):= 8192 *  4 +    1;
    MEMORY(8190):= 8192 *  9 + 8182;
    MEMORY(8191):= 8192 *  8 + 8177;
  end ESTABLISH_INITIAL_INSTRUCTIONS;

  procedure STORE_IN_MEMORY ( ITEM: INTEGER ) is
  begin
    MEMORY(AMEND_ADDRESS):= ITEM mod 262144;
    INCREMENT (AMEND_ADDRESS);
  end STORE_IN_MEMORY;

  procedure PREPARE_TO_JUMP ( TO: INTEGER ) is
  begin
--
-- 27-SEP-1997 This bit removed because it clutters the output
--
--    if MACHINE_STATE /= UNDEFINED_STATE then
--      MESSAGE_PUT ( "Machine Reset" );
--    end if;
--
    MEMORY(B_REGISTER_ADDRESS - 1):= SEQUENCE_CONTROL_REGISTER;
    B_REGISTER_ADDRESS := 1;
    SEQUENCE_CONTROL_REGISTER := TO;
--
--  05-APR-1998
--
    if TO = 18 then run_after_lib := 1;
               else run_after_lib := 0;
    end if;
--
    MACHINE_STATE := RUNNABLE_STATE;
  end PREPARE_TO_JUMP;

  procedure COMBINED_LEFT_SHIFT is
  begin
    if Q_REGISTER >= 131072 then
      ACCUMULATOR := (ACCUMULATOR * 2) mod 262144 + 1;
    else
      ACCUMULATOR := (ACCUMULATOR * 2) mod 262144;
    end if;
    Q_REGISTER := (Q_REGISTER * 2) mod 262144;
  end COMBINED_LEFT_SHIFT;

  procedure COMBINED_RIGHT_SHIFT is
  begin
    if ACCUMULATOR mod 2 = 1 then
      Q_REGISTER := Q_REGISTER / 2 + 131072;
    else
      Q_REGISTER := Q_REGISTER / 2;
    end if;
    if ACCUMULATOR >= 131072 then
      ACCUMULATOR := ACCUMULATOR / 2 + 131072;
    else
      ACCUMULATOR := ACCUMULATOR / 2;
    end if;
  end COMBINED_RIGHT_SHIFT;

  function ODD_PARITY return BOOLEAN is
    SHIFT: INTEGER := PERIPHERAL_DATA;
    TOTAL: BOOLEAN := FALSE;
  begin
    while SHIFT > 0 loop
      if SHIFT mod 2 = 1 then TOTAL := not TOTAL; end if;
      SHIFT := SHIFT / 2;
    end loop;
    return TOTAL;
  end ODD_PARITY;

begin --SIM900AL-NT-R

  CLEAR_STORE;
--  ESTABLISH_INITIAL_INSTRUCTIONS;

  MESSAGE_PUT("Version "&VERSION&" of "&DATE);

  open(file => MODULE, name => m0, mode => in_file);
  set_input(MODULE);
  set_line_length(72);
  for I in 0..8191 loop
    Ada.Integer_text_IO.get(top);   -- Use the new Integer-IO to get the File
    Ada.Integer_text_IO.get(bot);   -- into memory, EB, 12-3-2008
    OPERAND_ADDRESS := INTEGER(I);
    ACCUMULATOR := INTEGER(top) * 8192 +
                   INTEGER(bot);
    MEMORY(OPERAND_ADDRESS):= ACCUMULATOR;
  end loop;
  close(file => MODULE);

  open(file => MODULE, name => m1, mode => in_file);
  set_input(MODULE);
  for I in 8192..16383 loop
    Ada.Integer_text_IO.get(top);  -- Use the new Integer-IO to get the File
    Ada.Integer_text_IO.get(bot);  -- into memory, EB, 12-3-2008
    OPERAND_ADDRESS := INTEGER(I);
    ACCUMULATOR := INTEGER(top) * 8192 +
                   INTEGER(bot);
    MEMORY(OPERAND_ADDRESS):=ACCUMULATOR;
  end loop;
  close(file => MODULE);

  open(file => f, name => ein, mode => in_file);
  set_input(f);
  int_level := 1;

  SimStart:=Ada.Calendar.Clock;

  MESSAGE_PUT ( "Simulation Starting" );
  loop

    AMENDING_ON := FALSE;
    loop

      if STRING_MATCH ("KEYB")  then
        set_input(standard_input);
      elsif STRING_MATCH("EIN") then
        set_input(f);
      elsif STRING_MATCH("ADRLB") then
        OP := ADD_RLB;
--
-- 25-SEP-1997. New directive ADLIB.
--
      elsif STRING_MATCH("ADLIB") then
        lib_flag1 := TRUE;
      elsif STRING_MATCH ("IPBIN") then
        INPUT_MODE := BINARY_MODE;
      elsif INPUT_MODE = ASCII_MODE or else
      ( AMENDING_ON = FALSE and then CHARACTER_LOOK in '0'..'9' ) then
        if INPUT_MODE = ASCII_MODE then
          PERIPHERAL_DATA := CHARACTER'POS(CHARACTER_LOOK); CHARACTER_GET;
          if ODD_PARITY then
            PERIPHERAL_DATA := PERIPHERAL_DATA + 128;
          end if;
        else
          PERIPHERAL_DATA := NATURAL_GET;
          if PERIPHERAL_DATA >= 256 then
            ERROR_PUT ( "Bad Peripheral Data" );
          end if;
        end if;
        if MACHINE_STATE = INPUT_HUNG_STATE then
          exit;
        elsif PERIPHERAL_DATA = 0 then
          MESSAGE_PUT ( "Blank Ignored" );
        else
          ERROR_PUT ( "Character Ignored" );
        end if;
      elsif STRING_MATCH ("DUMP102") then
        -- Dump the virtual registers of simulated 102 if SIM is engaged.
        -- Otherwise the output is nonsense; 6-MAY-2011, EB.
        s102:=MEMORY(34)/64+MEMORY(35);
        a102:=MEMORY(32)/64;
        i102:=MEMORY(38)/64;
        i1021:=i102/256;
        i1022:=(i102-256*i1021)/128;
        i1023:=(i102-256*i1021)-128*i1022;
        b102:=MEMORY(33)/64;
        d102:=MEMORY(36)/8192+Memory(37)/128;
        MESSAGE_PUT("S="&Ada.Strings.Fixed.Trim(s102'img,Ada.Strings.Left)&
                  "; A="&Ada.Strings.Fixed.Trim(a102'img,Ada.Strings.Left)&
                  "; B="&Ada.Strings.Fixed.Trim(b102'img,Ada.Strings.Left)&
                  "; D="&Ada.Strings.Fixed.Trim(d102'img,Ada.Strings.Left)&
                  "; LastInstr="&Ada.Strings.Fixed.Trim(i1021'img&i1022'img,Ada.Strings.Left)&":"&Ada.Strings.Fixed.Trim(i1023'img,Ada.Strings.Left));
      elsif STRING_MATCH ("IPASC") then
        -- Without ignoring the following CRLF or anything before it.
        INPUT_MODE := ASCII_MODE;
      -- IPRAW added due to Terry's patch, 1-MAY-2010, EB
      elsif STRING_MATCH ("IPRAW") then
        if BYTE_IO.IS_OPEN (FILE) then BYTE_IO.CLOSE (FILE); end if;
        -- Perhaps should warn if significant characters unread?
        declare
          FILENAME: STRING(1..80) := (others=>' ');
        begin
           while CHARACTER_LOOK = ' ' loop CHARACTER_GET; end loop;
             for F in FILENAME'RANGE loop
                if CHARACTER_LOOK <= ' ' then
                    if F > 1 then
                       BYTE_IO.OPEN (FILE, BYTE_IO.IN_FILE, FILENAME(1..F-1));
                       MESSAGE_PUT("Loaded tape "&FILENAME(1..F-1)&" into reader.");
                    end if;
                    exit;
                end if;
                FILENAME(F) := CHARACTER_LOOK; CHARACTER_GET;
             end loop;
           exception when others =>
           ERROR_PUT ( "File Not Found" );
        end;
      -- Patch for writing tapes to binary files, 4-MAY-2010, EB.
      elsif STRING_MATCH ("CTAPE") then
        if BYTE_IO.IS_OPEN (TAPE) then BYTE_IO.CLOSE (TAPE); end if;
        declare
          FILENAME: STRING(1..80) := (others=>' ');
        begin
           while CHARACTER_LOOK = ' ' loop CHARACTER_GET; end loop;
             for F in FILENAME'RANGE loop
                if CHARACTER_LOOK <= ' ' then
                    if F > 1 then
                       BYTE_IO.CREATE (TAPE, BYTE_IO.OUT_FILE, FILENAME(1..F-1));
                       MESSAGE_PUT("Punching to file "&FILENAME(1..F-1)&".");
                    end if;
                    exit;
                end if;
                FILENAME(F) := CHARACTER_LOOK; CHARACTER_GET;
             end loop;
           exception when others =>
           ERROR_PUT ( "Error opening Tape-File" );
        end;

      elsif CHARACTER_LOOK <= ' ' then CHARACTER_GET;
      elsif CHARACTER_MATCH ('(') then IGNORE_COMMENTS;

      elsif CHARACTER_MATCH ('*') then
        CLEAR_STORE ( ( NATURAL_GET / 8192 ) * 8192 - 1 );
      elsif CHARACTER_MATCH ('^') then
        AMENDING_ON := TRUE;
        AMEND_ADDRESS := NATURAL_GET;
      elsif CHARACTER_MATCH ('>') then
        AMENDING_ON := TRUE;
        AMEND_ADDRESS := AMEND_ADDRESS + NATURAL_GET;
      elsif CHARACTER_MATCH ('<') then
        PREPARE_TO_JUMP (AMEND_ADDRESS);
      elsif CHARACTER_MATCH ('&') then
        STORE_IN_MEMORY ( NATURAL_GET (BASE=>8) );
      elsif CHARACTER_MATCH ('+') then
        STORE_IN_MEMORY ( NATURAL_GET );
      elsif CHARACTER_MATCH ('-') then
        STORE_IN_MEMORY ( 262144 - NATURAL_GET );

      elsif CHARACTER_LOOK = '/' or else CHARACTER_LOOK in '0'..'9' then
        if CHARACTER_MATCH ('/') then AMEND_WORD := 131072;
        else AMEND_WORD := 0; end if;
        AMEND_FIELD := NATURAL_GET;
        if AMEND_FIELD >= 16 then ERROR_PUT ( "Bad Function" ); end if;
        AMEND_WORD := AMEND_WORD + ( AMEND_FIELD mod 16 ) * 8192;
        while CHARACTER_LOOK <= ' ' loop CHARACTER_GET; end loop;
        if CHARACTER_MATCH (';') then
          if CHARACTER_MATCH ('+') then
            AMEND_FIELD := AMEND_ADDRESS + NATURAL_GET;
          elsif CHARACTER_MATCH ('-') then
            AMEND_FIELD := AMEND_ADDRESS - NATURAL_GET;
          else ERROR_PUT ( "Bad Address" ); end if;
        else
          AMEND_FIELD := NATURAL_GET;
        end if;
        STORE_IN_MEMORY ( AMEND_WORD + AMEND_FIELD mod 8192 );

      elsif STRING_MATCH ("LOAD") then
        ESTABLISH_INITIAL_INSTRUCTIONS;
        PREPARE_TO_JUMP (8181); exit;
      elsif STRING_MATCH ("JUMP") then
        PREPARE_TO_JUMP ( PARAMETER_GET ); exit;
      elsif STRING_MATCH ("GO") then
        if MACHINE_STATE = UNDEFINED_STATE then
          ERROR_PUT ( "No Continuation" );
        else -- Have no PERIPHERAL_DATA so must cancel any INPUT_HUNG_STATE.
          MACHINE_STATE := RUNNABLE_STATE; exit;
        end if;
      elsif STRING_MATCH ("QUIT") then
        PREPARE_TO_JUMP (0); -- Just to check MACHINE_STATE.
        MESSAGE_PUT ( "Simulation Completed after"&InstCount'img&" instructions," );
        SimStop:=Ada.Calendar.Clock;
        Split(SimStart, Year, Month, Day, Start);
        Split(SimStop, Year, Month, Day, Stop);
        Stop:=Stop-Start;
        MESSAGE_PUT ("time elapsed:"&Stop'img&"s on "&Ada.Strings.Fixed.Trim(Day'img,Ada.Strings.Left)&"-"&Month_Name(Month)&"-"&Ada.Strings.Fixed.Trim(Year'img,Ada.Strings.Left)&".");
        return;

      elsif STRING_MATCH ("OPASC") then
        OUTPUT_MODE := ASCII_MODE;
      elsif STRING_MATCH ("OPBIN") then
        OUTPUT_MODE := BINARY_MODE;
      elsif STRING_MATCH ("OPLEG") then
        OUTPUT_MODE := LEGIBLE_MODE;
      elsif STRING_MATCH ("OPSIN") then
        OUTPUT_MODE := SINK_MODE;
      elsif STRING_MATCH ("COPY") then
        while CHARACTER_MATCH (' ') loop null; end loop;
        ENSURE_NEW_LINE;
        while CHARACTER_LOOK /= CRLF loop
          CHARACTER_PUT ( CHARACTER_LOOK ); CHARACTER_GET;
        end loop;
        ENSURE_NEW_LINE;

      elsif STRING_MATCH ("DISPLAY") then
        DISPLAY_FROM := PARAMETER_GET; DISPLAY_TO := PARAMETER_GET;
        for DISPLAY_ADDRESS in DISPLAY_FROM..DISPLAY_TO loop
          INFORMATION_PUT ( DISPLAY_ADDRESS,
            DISPLAY_REGISTERS_RATHER_THAN_FORMATS => FALSE );
        end loop;
      elsif STRING_MATCH ("TRACE") then
        TRACING_ON := TRUE;
      elsif STRING_MATCH ("TROFF") then
        TRACING_ON := FALSE;
      elsif STRING_MATCH ("BREAK") then
        BREAKPOINT_ADDRESS := PARAMETER_GET;
      elsif STRING_MATCH ("BROFF") then
        BREAKPOINT_ADDRESS := -1;
      elsif STRING_MATCH ("BATCH") then
        BATCH_MODE_RATHER_THAN_INTERACTIVE := TRUE;
      elsif STRING_MATCH ("INTER") then
        BATCH_MODE_RATHER_THAN_INTERACTIVE := FALSE;

      else
        ERROR_PUT ( "Unknown Command" );
        CHARACTER_GET;
      end if;

    end loop;

    ------------------ Simulation loop -----------------

    loop

      ----------------------------------------------------------------------
      -- Count Instructions and every 1000 instructions check for events or
      -- the necessity to redraw the graphics window due to plotter commands
      InstCount:= InstCount+1 ;
      if InstCount mod 1000 = 0 then
        -------------------------- GUI events handler ----------------------
        if Gtk.Main.Events_Pending then
                                       DummyMain:= Gtk.Main.Main_Iteration;
                                   end if;
        if GUIquit then
                       raise FATAL_EXCEPTION;
                   end if;
        --- Redraw window if plotter instructions have occured -------------
        if plot then
             if PicChanged then
                  Draw(Window);
                  PicChanged:=False;
                  end if;
              end if;
        end if;
      ------------------------------------------------------EB,14-MAR-2008--

      INSTRUCTION := MEMORY (SEQUENCE_CONTROL_REGISTER);

      OPERAND_ADDRESS := (INSTRUCTION mod 8192) +
        ( (SEQUENCE_CONTROL_REGISTER / 8192) * 8192 );
      if INSTRUCTION >= 131072 then
-- 21-APR-1996
        Q_REGISTER := 0;
-- 11-MAY-1996. Should be mod 131072 but 65536 covers up for a bit of
--              bad coding in Call By Name.
        OPERAND_ADDRESS :=
       ( OPERAND_ADDRESS + MEMORY (B_REGISTER_ADDRESS) ) mod 65536;

      end if;

      OLD_SEQUENCE_CONTROL_REGISTER := SEQUENCE_CONTROL_REGISTER;
      SEQUENCE_CONTROL_REGISTER := SEQUENCE_CONTROL_REGISTER + 1;

      case (INSTRUCTION / 8192) mod 16 is

      when 0 => -- Set B-register.
        Q_REGISTER := MEMORY (OPERAND_ADDRESS);
        MEMORY(B_REGISTER_ADDRESS):= Q_REGISTER;

      when 1 => -- Add.
        ACCUMULATOR :=
          ( MEMORY (OPERAND_ADDRESS) + ACCUMULATOR ) mod 262144;

      when 2 => -- Negate & add.
        Q_REGISTER := MEMORY (OPERAND_ADDRESS);
        ACCUMULATOR := (Q_REGISTER - ACCUMULATOR) mod 262144;

      when 3 => -- Store Q-register.
        MEMORY(OPERAND_ADDRESS):= Q_REGISTER / 2;

      when 4 => -- Read.
        ACCUMULATOR := MEMORY (OPERAND_ADDRESS);

      when 5 => -- Write.
        MEMORY(OPERAND_ADDRESS):=ACCUMULATOR;

         when 6 => -- Collate.
          --  For test purposes, 20-MAR-2008, EB
          --  MESSAGE_PUT("A"&ACCUMULATOR'Img&"A"&MEMORY (OPERAND_ADDRESS)'Img&"=");
          Bit_And ( MEMORY (OPERAND_ADDRESS)'Address,32, ACCUMULATOR'Address,32, ACCUMULATOR'Address);
          --  MESSAGE_PUT("="&ACCUMULATOR'Img&"!");
      when 7 => -- Jump if zero.
        if ACCUMULATOR = 0 then
          SEQUENCE_CONTROL_REGISTER := OPERAND_ADDRESS;
        end if;

      when 8 => -- Jump.
        SEQUENCE_CONTROL_REGISTER := OPERAND_ADDRESS;

      when 9 => -- Jump if negative.
        if ACCUMULATOR >= 131072 then
          SEQUENCE_CONTROL_REGISTER := OPERAND_ADDRESS;
        end if;

      when 10 => -- Count in store.
        MEMORY (OPERAND_ADDRESS) :=
          ( MEMORY (OPERAND_ADDRESS) + 1 ) mod 262144;

      when 11 => -- Store SCR.
        Q_REGISTER := (SEQUENCE_CONTROL_REGISTER / 8192) * 8192;
        MEMORY (OPERAND_ADDRESS):= SEQUENCE_CONTROL_REGISTER mod 8192;

      when 12 => -- Multiply.
        MEMORY_BUFFER := MEMORY (OPERAND_ADDRESS);
-- 01-FEB-2011 removed Don's code an included Terry's recommendation
          for PROCESS_COUNTER in 1..18 loop
            if PROCESS_COUNTER = 1 then
              Q_REGISTER := ACCUMULATOR; ACCUMULATOR := 0;
              X_BITS := Q_REGISTER * 2;
            else
              X_BITS := Q_REGISTER;
              COMBINED_RIGHT_SHIFT;
            end if;
            case X_BITS mod 4 is
            when 1 =>
              ACCUMULATOR := (ACCUMULATOR + MEMORY_BUFFER) mod 262144;
            when 2 =>
              ACCUMULATOR := (ACCUMULATOR - MEMORY_BUFFER) mod 262144;
            when others =>
              null;
            end case;
          end loop;
          if MEMORY_BUFFER = 131072 and then ACCUMULATOR /=0 then
             -- Correct for using 18 bits where hardware uses 19 bits.
             ACCUMULATOR := 262144 - ACCUMULATOR;
          end if;
-- 01-FEB-2011, end of new multiply.

      when 13 => -- Divide.
        MEMORY_BUFFER := MEMORY (OPERAND_ADDRESS);
        Q_REGISTER := (Q_REGISTER / 2) * 2;
        X_BITS := ACCUMULATOR;
        for PROCESS_COUNTER in 1..18 loop
          if ( X_BITS >= 131072 ) = ( MEMORY_BUFFER >= 131072 ) then
            Q_REGISTER := Q_REGISTER + 1;
            ACCUMULATOR := (ACCUMULATOR - MEMORY_BUFFER) mod 262144;
          else
            ACCUMULATOR := (ACCUMULATOR + MEMORY_BUFFER) mod 262144;
          end if;
          X_BITS := ACCUMULATOR;
          COMBINED_LEFT_SHIFT;
        end loop;
        ACCUMULATOR := Q_REGISTER + 1;

      when 14 => -- Shift, etc..
        OPERAND_ADDRESS := OPERAND_ADDRESS mod 8192;
        if OPERAND_ADDRESS < 4096
        then -- Left shift.
          for PROCESS_COUNTER in 1..OPERAND_ADDRESS loop
            COMBINED_LEFT_SHIFT;
          end loop;
        else -- Right shift.
          for PROCESS_COUNTER in OPERAND_ADDRESS..8191 loop
            COMBINED_RIGHT_SHIFT;
          end loop;
        end if;

      when 15 => -- Input/output, etc..
        OPERAND_ADDRESS := OPERAND_ADDRESS mod 8192;
        if OPERAND_ADDRESS = 7168
        then -- Level terminate.
          if int_level < 4 then
            MEMORY(B_REGISTER_ADDRESS - 1):= SEQUENCE_CONTROL_REGISTER;
            B_REGISTER_ADDRESS := 7;
            SEQUENCE_CONTROL_REGISTER := MEMORY (B_REGISTER_ADDRESS - 1);
            int_level := int_level + 1;
          else
            MESSAGE_PUT ( "Interrupt level 4 already reached!" );
            B_REGISTER_ADDRESS := 7;
            SEQUENCE_CONTROL_REGISTER := MEMORY (B_REGISTER_ADDRESS - 1);
            -- MESSAGE_PUT ( "Interrupt level 4 not to be exceeded! Dumping:" );
            -- raise TROUBLE;
          end if;
        -- Don Hunters Patch from 09-JAN-2003, retrofitted, but
        -- completely rewritten for Gnat-Ada, 8-OCT-2010, EB
        elsif OPERAND_ADDRESS = 123 then
          SimStop:=Ada.Calendar.Clock;
          Split(SimStop, Year, Month, Day, Stop);
          Hour:=Integer(Stop)/3600;
          Minute:=Integer(Stop)/60-60*Hour;
          Second:=Integer(Stop)-60*Minute-3600*Hour;
          MESSAGE_PUT ("time stamp:"&Hour'img&"h"&Minute'img&"m"&Second'img&"s :"&Day'img&" - "&Month_Name(Month)&" -"&Year'img&"");
        elsif OPERAND_ADDRESS < 4096
        then -- Input -- Modification: Raw input added, Terry's patch from 1-MAY-2010
           if BYTE_IO.IS_OPEN (FILE)
          and then not BYTE_IO.END_OF_FILE (FILE) then
            BYTE_IO.READ (FILE, BYTE);
            -- Changed TJF's LONIG_INTEGER to INTEGER, 3-MAY-2010, EB
            PERIPHERAL_DATA := INTEGER (BYTE);
            if ACCUMULATOR mod 2 = 1 then
              PERIPHERAL_DATA := PERIPHERAL_DATA mod 128;
            else
              PERIPHERAL_DATA := PERIPHERAL_DATA mod 256;
            end if;
          elsif MACHINE_STATE = RUNNABLE_STATE then
            SEQUENCE_CONTROL_REGISTER := OLD_SEQUENCE_CONTROL_REGISTER;
            -- End of patch
            MACHINE_STATE := INPUT_HUNG_STATE; exit;
          elsif ACCUMULATOR mod 2 = 1 then
            PERIPHERAL_DATA := PERIPHERAL_DATA mod 128;
          else
            PERIPHERAL_DATA := PERIPHERAL_DATA mod 256;
          end if;
          -- MACHINE_STATE = INPUT_HUNG_STATE.
          ACCUMULATOR := ( ACCUMULATOR * 128 + PERIPHERAL_DATA ) mod 262144;
          MACHINE_STATE := RUNNABLE_STATE;
        else -- Output.
          PERIPHERAL_DATA := ACCUMULATOR mod 256;
          -- MESSAGE_PUT("Output "&PERIPHERAL_DATA'Img&" -> "&OPERAND_ADDRESS'Img&".");
          case OUTPUT_MODE is
          when BINARY_MODE =>
            if OUTPUT_COLUMN > 60 then CHARACTER_PUT (CRLF);
            elsif OUTPUT_COLUMN /= 0 then CHARACTER_PUT (' ');
            end if;
            NATURAL_PUT (PERIPHERAL_DATA);
          when ASCII_MODE =>
-- 10-SEP-2000, plotter code, EB
            if OPERAND_ADDRESS = 4865 then
                  x := integer(ACCUMULATOR); x1 := x; x2 := x;
            elsif OPERAND_ADDRESS = 4866 then
                  y := integer(ACCUMULATOR); y1 := y; y2 := y;
            elsif OPERAND_ADDRESS = 4864 then
               if not plot then
                  plot := true;
                  -- Initialize and show "Elliott" window, 14-MAR-2008, EB
                  Gtk.Main.Init;
                  Gtk_New (Window, Window_Toplevel);
                  Set_Title (Window, "SIM900AL-NT-R - Elliott - 18bit Simulator");
                  Set_Border_Width (Window, Border_Width => 5);
                  Set_Policy (Window,
                      Allow_Shrink => False,
                      Allow_Grow   => False,
                      Auto_Shrink  => False);
                  Destroyed.Connect (Window, "destroy", Destroyed.To_Marshaller (Bye'Access));
                  Gtk_New_Vbox (Vbox, Homogeneous => False, Spacing => 0);
                  Add (Window, Vbox);
                  Show (Vbox);
                  Gtk_New(Drawing_Area);
                  Size (Drawing_Area, 640, 480);
                  Pack_Start (In_Box=>Vbox, Child=>Drawing_Area);
                  Show (Drawing_Area);
                  Unrealize (Drawing_Area);
                  Set_Events (Drawing_Area, Exposure_Mask or Leave_Notify_Mask or
                       Button_Press_Mask or Pointer_Motion_Mask or
                       Pointer_Motion_Hint_Mask);
                  Configured.Connect (Widget =>Drawing_Area,
                       Name   => "expose_event",
                       Marsh  => Configured.To_Marshaller (Expose_Event'Access));
                  Configured.Connect (Widget =>Drawing_Area,
                       Name   => "motion_notify_event",
                       Marsh  => Configured.To_Marshaller (Motion_Notify_Event'Access));
                  Configured.Connect (Widget =>Drawing_Area,
                       Name   => "configure_event",
                       Marsh  => Configured.To_Marshaller (Configure_Event'Access));
                  Configured.Connect (Widget =>Drawing_Area,
                       Name   => "button_press_event",
                       Marsh  => Configured.To_Marshaller (Button_Press_Event'Access));
                  Gtk_New (Button, "Kill Elliott");
                  Pack_Start (Vbox, Button, Expand=>False, Fill=>False);
                  Destroyed.Object_Connect (Button, "clicked", Destroyed.To_Marshaller (Bye'Access),
                       Slot_Object => Window);
                  Show_All (Window);
                  PicChanged:=True;   -- Request redraw
                  -- Window now online and ready for drawing...
               end if;
               X_Bits:=1;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  x2 := x+1;
               end if;
               X_Bits:=2;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  x2 := x-1;
               end if;
               X_Bits:=4;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  y2 := y-1;
               end if;
               X_Bits:=8;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  y2 := y+1;
               end if;
               X_Bits:=16;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  down := false;
               end if;
               X_Bits:=32;
               Bit_And ( PERIPHERAL_DATA'Address,32, X_Bits'Address,32, X_Bits'Address);
               if X_Bits /= 0 then
                  down := true;
               end if;
                  x1 := x;  y1 := y;
                  x  := x2;  y := y2;
               if down then
                     -- Do the Graphics work via GtkAda, 14-MAR-2008, EB
                     Draw_Line(Pixmap,Get_Black(Get_Style(Window)),Gint(x1),Gint(y1),Gint(x2),Gint(y2));
                     PicChanged:=True;  -- Request redraw of PixMap
               end if;
            elsif (OPERAND_ADDRESS) = 6144 and (BYTE_IO.IS_OPEN (TAPE)) then
               BYTE:=BYTE_INTEGER(PERIPHERAL_DATA);
               BYTE_IO.WRITE(TAPE,BYTE);
               -- For testing of the tape trap routine, 4-MAY-2010, EB
               -- MESSAGE_PUT("Output "&PERIPHERAL_DATA'Img&" -> "&OPERAND_ADDRESS'Img&".");
            else
-- end of plotter code, EB.
               if ODD_PARITY then ERROR_PUT ( "Parity Error" ); end if;
               PERIPHERAL_DATA := PERIPHERAL_DATA mod 128;
               case PERIPHERAL_DATA is
               when 0 | 13 | 127 => null;
               when others => CHARACTER_PUT ( CHARACTER'VAL (PERIPHERAL_DATA) );
               end case;
            end if;
          when LEGIBLE_MODE =>
            ENSURE_NEW_LINE;
            for I in 1..8 loop
              if PERIPHERAL_DATA >= 128
              then CHARACTER_PUT ('O'); else CHARACTER_PUT ('.'); end if;
              if I = 5 then CHARACTER_PUT ('o'); end if;
              PERIPHERAL_DATA := (PERIPHERAL_DATA * 2) mod 256;
            end loop;
            ENSURE_NEW_LINE;
          when others => null; end case; -- SINK_MODE.
        end if;

      when others => null; end case; -- Impossible.

      if TRACING_ON then
        INFORMATION_PUT ( OLD_SEQUENCE_CONTROL_REGISTER,
          DISPLAY_REGISTERS_RATHER_THAN_FORMATS => TRUE );
      end if;

--
--  This whole section has changed on 27-SEP-1997.
--

      if SEQUENCE_CONTROL_REGISTER = OLD_SEQUENCE_CONTROL_REGISTER
      and SEQUENCE_CONTROL_REGISTER /= 7943 then
         if SEQUENCE_CONTROL_REGISTER = 1727
         and (lib_flag1 or (OP = ADD_RLB))
         and not has_run then
            INPUT_MODE := BINARY_MODE;
            PREPARE_TO_JUMP(9);
            if lib_flag1 then
               open(file => lib, name => library, mode => in_file);
               set_input(lib);
               lib_flag1 := FALSE;
               lib_flag2 := TRUE;
            else
               OP := TRANSLATE;
            end if;
         else
            MACHINE_STATE := UNDEFINED_STATE;
            CHARACTER_PUT(CRLF); CHARACTER_PUT(CRLF);
            if not has_run then
               MESSAGE_PUT("Warning: No ADLIB directive - Algol won't run");
               -- Show message only once, 2-SEP-2010, EB
               has_run := True;
            end if;
            MESSAGE_PUT("Stop at "&SEQUENCE_CONTROL_REGISTER'img&" encountered");
            -- Included above: (Nicer output, 1-MAY-2010, EB) WORD_PUT(' ', SEQUENCE_CONTROL_REGISTER);
            exit;
         end if;
      elsif SEQUENCE_CONTROL_REGISTER = 7943 then
         if OP = TRANSLATE then
            INPUT_MODE := ASCII_MODE;
--
            if run_after_lib = 1 then
               has_run := FALSE;
               OP := TRANSLATE;
               run_after_lib := 0;
               PREPARE_TO_JUMP(8);
            elsif run_after_lib = 0 then
--
               OP := RUN;
               if lib_flag2 then
                  close(file => lib);
                  set_input(f);
                  lib_flag2 := FALSE;
               end if;
               has_run := TRUE;
               PREPARE_TO_JUMP(10);
            end if;
         elsif OP = ADD_RLB then
            INPUT_MODE := BINARY_MODE;
            OP := TRANSLATE;
            if lib_flag2 then
               close(file => lib);
               set_input(f);
               lib_flag2 := FALSE;
            end if;
            PREPARE_TO_JUMP(14);
         end if;
      elsif SEQUENCE_CONTROL_REGISTER = BREAKPOINT_ADDRESS then
        MESSAGE_PUT ( "Hit Breakpoint" ); exit;
      end if;

    end loop;

  end loop;

  close(file => f);

  -- Close open file (Terry's patch 1-MAY-2010, moved here)
  if BYTE_IO.IS_OPEN (FILE) then BYTE_IO.CLOSE (FILE); end if;
  -- Perhaps should warn if significant characters unread?

  -- Close eventally open tape output file, 4-MAY-2010, EB
  if BYTE_IO.IS_OPEN (TAPE) then BYTE_IO.CLOSE (TAPE); end if;

  -- Waiting for GUI to close the still open window, 3-MAR-2008, EB
  if plot then
         Gtk.Main.Main;
         plot:=False;
  end if;


  exception
  when TROUBLE =>
     for I in 0..7 loop
        OPERAND_ADDRESS := INTEGER(I);
        INFORMATION_PUT(OPERAND_ADDRESS,
            DISPLAY_REGISTERS_RATHER_THAN_FORMATS => FALSE);
     end loop;
     INFORMATION_PUT(135, TRUE);
     MESSAGE_PUT ( "Simulation EMERGENCY stopped after"&InstCount'img&" instructions," );
        SimStop:=Ada.Calendar.Clock;
        Split(SimStart, Year, Month, Day, Start);
        Split(SimStop, Year, Month, Day, Stop);
        Stop:=Stop-Start;
        MESSAGE_PUT ("time elapsed:"&Stop'img&"s on "&Ada.Strings.Fixed.Trim(Day'img,Ada.Strings.Left)&"-"&Month_Name(Month)&"-"&Ada.Strings.Fixed.Trim(Year'img,Ada.Strings.Left)&".");
     return;

  when NAME_ERROR =>
     MESSAGE_PUT("Either ElliottIn.dat (Input File), one of the memory files");
     MESSAGE_PUT("Sim900al-NT.M0.dat or Sim900al-NT.M1.dat or a valid library");
     MESSAGE_PUT("Sim900al-NT.LIBRARY.dat is missing. Provide the missing");
     MESSAGE_PUT("File(s) and try again.");
    return;

  when CONSTRAINT_ERROR =>
--      iio.put(x1); iio.put(y1); iio.put(x2); iio.put(y2);
        if x2 < 0 then MESSAGE_PUT("Off to left of screen"); end if;
        if y2 < 0 then MESSAGE_PUT("Off above screen"); end if;
--      INFORMATION_PUT(ACCUMULATOR,
--          DISPLAY_REGISTERS_RATHER_THAN_FORMATS => FALSE);
     return;

   when fatal_exception =>
        PREPARE_TO_JUMP (0); -- Just to check MACHINE_STATE.
        -- Potentially print count of simulated instructions, 14-MAR-2008, EB
        -- Ada.Long_Integer_Text_IO.Put(InstCount); -- Added below, 15-MAR-2008, EB
        -- Waiting for GUI to close the still open window
        if plot then
             if not GUIquit then WAITquit:= True; Gtk.Main.Main; end if;
             Plot:=False;
        end if;
        MESSAGE_PUT ( "Simulation Completed after"&InstCount'img&" instructions," );
        SimStop:=Ada.Calendar.Clock;
        Split(SimStart, Year, Month, Day, Start);
        Split(SimStop, Year, Month, Day, Stop);
        Stop:=Stop-Start;
        MESSAGE_PUT ("time elapsed:"&Stop'img&"s on "&Ada.Strings.Fixed.Trim(Day'img,Ada.Strings.Left)&"-"&Month_Name(Month)&"-"&Ada.Strings.Fixed.Trim(Year'img,Ada.Strings.Left)&".");
        return;

end SIM900AL_NT_R;

-- Add ENTER and OBEY ?
-- Add 3 Interrupts (and "hardware" trace) ?
-- Add fuller Strobe & Trap on Instruction or Operand or Peripheral ?
-- Separate out the Teleprinter & Paper Tape channels, & add Echoing ?
-- Add the single word & block input/output ?
-- Add 905/920C extra instructions & addressing modes ?
-- Add machine-distinguishing side-effects ?
-- Add an INCLUDE and an IPASC_INCLUDE_IPBIN facility ?
-- Make prompting when interactive appropriate for MACHINE_STATE ?
-- Add stepping by a given number of instructions ?
-- Add timing analysis ?
-- Do proper checking of Jumping & Amending Addresses ?
-- Make memory size bigger or dynamic ?
-- Unroll the Collate, Multiply, & Divide loops ?
-- Use a case for Shift instructions ?
-- Invoke OPLEG automatically somehow ?
